#include <bits/stdc++.h>

using namespace std;
using ll = long long;

constexpr int N = 1e5 + 7, B_sz = 450, B_cnt = N / B_sz + 1;

namespace CoolAdder {
    int blockSum[B_cnt], val[N], orderPosition[N];

    void clear() {
        fill(blockSum, blockSum + B_cnt, 0);
        fill(val, val + N, 0);
    }

    void add(int i, int x) {
        i = orderPosition[i];
        blockSum[i / B_sz] += x;
        val[i] += x;
    }

    int query(int i) { // excluding i
//        i = orderPosition[i];
        int b = i / B_sz;
        int ans = 0;
        for (int j = 0; j < b; ++j) {
            ans += blockSum[j];
        }
        for (int j = b * B_sz; j < i; ++j) {
            ans += val[j];
        }
        return ans;
    }
}

int n, q;

int blockType[B_cnt]; // 0 or 1
int addons[B_cnt][N];
vector<int> notMine[B_cnt];
int type[N]; // type[i] is equal to -1 if it is not in "notMine"

void putPoint(int i, int v) {
    if (type[i] == v) {
        return;
    }
    int b = i / B_sz;
    if (type[i] == -1) {
        notMine[b].push_back(i);
        CoolAdder::add(i, v - blockType[b]);
        type[i] = v;
    } else {
        CoolAdder::add(i, v - type[i]);
        type[i] = v;
    }
}

void putBlock(int b, int v) {
    for (int u : notMine[b]) {
        CoolAdder::add(u, (blockType[b] - type[u]));
        type[u] = -1;
    }
    notMine[b].clear();
    blockType[b] = v;
}

void putSegment(int l, int r, int v) { // inclusive
    int bl = l / B_sz, br = r / B_sz;
    if (bl == br) {
        for (int i = l; i <= r; ++i) {
            putPoint(i, v);
        }
        return;
    }
    for (int i = l; i < (bl + 1) * B_sz; ++i) {
        putPoint(i, v);
    }
    for (int b = bl + 1; b < br; ++b) {
        putBlock(b, v);
    }
    for (int i = br * B_sz; i <= r; ++i) {
        putPoint(i, v);
    }
}

int query(int i) {
    int ans = CoolAdder::query(i);
    for (int b = 0; b < B_cnt; ++b) {
        ans += addons[b][i] * blockType[b];
    }
    return ans;
}

struct Query {
    int type, l, r, x;
};

constexpr int maxQ = 1e5 + 1;

Query queries[maxQ];

vector<int> solve(vector<int> ord, vector<int> T, int tp) {
    CoolAdder::clear();
    for (int i = 0; i < n; ++i) {
        CoolAdder::orderPosition[ord[i]] = i;
    }
    struct SegmentTree {
        vector<int> t;
        int sz = 1;

        void init(int m) {
            sz = m;
            t.assign(sz * 2, -1);
        }

        void rangeApply(int l, int r, int v) {
            for (l += sz, r += sz; l < r; l >>= 1, r >>= 1) {
                if (l & 1) {
                    t[l] = max(t[l], v);
                    l += 1;
                }
                if (r & 1) {
                    r -= 1;
                    t[r] = max(t[r], v);
                }
            }
        }

        int query(int x) {
            int ans = -1;
            for (x += sz; x > 0; x >>= 1) {
                ans = max(ans, t[x]);
            }
            return ans;
        }
    };
    SegmentTree st;
    st.init(n);
    fill(type, type + n, -1);
    fill(blockType, blockType + B_cnt, 0);
    for (int i = 0; i < B_cnt; ++i) {
        int l = i * B_sz, r = min(n, (i + 1) * B_sz);
        fill(addons[i], addons[i] + n, 0);
        blockType[i] = (tp == -1);
        notMine[i].clear();
        if (l < r) {
            for (int j = 0, sum = 0; j < n; ++j) {
                addons[i][j] = sum;
                sum += ord[j] >= l && ord[j] < r;
            }
        }
    }
    for (int i = 0; i < B_cnt; ++i) {
        int l = i * B_sz, r = min(n, (i + 1) * B_sz);
        if (l < r) {
            putBlock(i, (-1 == tp));
        }
    }
    vector<int> answers(q);
    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 1) {
            int l = queries[i].l, r = queries[i].r, t = queries[i].x == tp;
            putSegment(l, r, t);
            st.rangeApply(l, r + 1, i);
        } else {
            answers[i] = query(T[i]);
            int vert = ord[T[i]];
            int fx = st.query(vert);
            fx = (fx == -1 ? -1 : queries[fx].x);
            answers[i] += (tp == fx);
        }
    }
    return answers;
}

struct SegmentTree {
    vector<int> t;
    int sz = 1;

    void init(int m) {
        sz = m;
        t.assign(sz * 2, -1);
    }

    void rangeApply(int l, int r, int v) {
        for (l += sz, r += sz; l < r; l >>= 1, r >>= 1) {
            if (l & 1) {
                t[l] = max(t[l], v);
                l += 1;
            }
            if (r & 1) {
                r -= 1;
                t[r] = max(t[r], v);
            }
        }
    }

    int query(int x) {
        int ans = -1;
        for (x += sz; x > 0; x >>= 1) {
            ans = max(ans, t[x]);
        }
        return ans;
    }
};

vector<int> stupid(vector<array<int, 2>> ch) {
    vector<int> t(n, -1), ans(q);
    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 1) {
            for (int x = queries[i].l; x <= queries[i].r; ++x) {
                t[x] = queries[i].x;
            }
        } else {
            vector<int> ord;
            auto dfs = [&](auto self, int v) -> void {
                if (t[v] == -1) {
                    ord.push_back(v);
                }
                if (ch[v][0] != -1) {
                    self(self, ch[v][0]);
                }
                if (t[v] == 0) {
                    ord.push_back(v);
                }
                if (ch[v][1] != -1) {
                    self(self, ch[v][1]);
                }
                if (t[v] == 1) {
                    ord.push_back(v);
                }
            };
            dfs(dfs, 0);
            ans[i] = find(ord.begin(), ord.end(), queries[i].l) - ord.begin() + 1;
        }
    }
    return ans;
}

vector<int> smart(vector<array<int, 2>> ch) {
    vector<int> L, Mid, R;
    vector<int> tL(n), sz(n), depth(n), init(n), tMid(n), tR(n);
    auto dfs = [&](auto self, int v) -> void {
        tL[v] = size(L);
        L.push_back(v);
        if (ch[v][0] != -1) {
            depth[ch[v][0]] = depth[v] + 1;
            init[ch[v][0]] = init[v];
            self(self, ch[v][0]);
            sz[v] += sz[ch[v][0]];
        }
        tMid[v] = size(Mid);
        Mid.push_back(v);
        if (ch[v][1] != -1) {
            depth[ch[v][1]] = depth[v] + 1;
            init[ch[v][1]] = init[v] + (ch[v][0] == -1 ? 0 : sz[ch[v][0]]);
            self(self, ch[v][1]);
            sz[v] += sz[ch[v][1]];
        }
        tR[v] = size(R);
        sz[v] += 1;
        R.push_back(v);
    };
    dfs(dfs, 0);

    vector<int> answers(q), tmp, T(q);

    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 2) {
            int x = queries[i].l;
            T[i] = tL[x] + sz[x] - 1;
        }
    }
    tmp = solve(L, T, -1);
    for (int i = 0; i < q; ++i) {
        answers[i] += tmp[i];
    }
//    for (int x : answers) {
//        cout << x << " ";
//    }
//    cout << endl;

    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 2) {
            int x = queries[i].l;
            T[i] = tR[x];
        }
    }
    tmp = solve(R, T, -1);
    for (int i = 0; i < q; ++i) {
        answers[i] -= tmp[i];
    }

//    for (int x : answers) {
//        cout << x << " ";
//    }
//    cout << endl;

    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 2) {
            int x = queries[i].l;
            T[i] = tMid[x] + (ch[x][1] == 0 ? 0 : sz[ch[x][1]]);
        }
    }
    tmp = solve(Mid, T, 0);
    for (int i = 0; i < q; ++i) {
        answers[i] += tmp[i];
    }

//    for (int x : answers) {
//        cout << x << " ";
//    }
//    cout << endl;

    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 2) {
            int x = queries[i].l;
            T[i] = tR[x];
        }
    }
    tmp = solve(R, T, 0);
    for (int i = 0; i < q; ++i) {
        answers[i] -= tmp[i];
    }

//    for (int x : answers) {
//        cout << x << " ";
//    }
//    cout << endl;

    SegmentTree st;
    st.init(n);

    vector<int> ans(q);

    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 2) {
            int x = queries[i].l;
            int fx = st.query(x);
            if (fx != -1) {
                fx = queries[fx].x;
            }
//            cerr << fx << " " << init[x] << endl;
            ans[i] = answers[i] + init[x] + (fx == -1 ? 0 : fx == 0 ? (ch[x][0] == -1 ? 0 : sz[ch[x][0]]) : sz[x] - 1) + 1;
        } else {
            st.rangeApply(queries[i].l, queries[i].r + 1, i);
        }
    }

    return ans;
}

void stress() {
    mt19937 rnd(228);
    while (true) {
        n = rnd() % 2 + 2;
        q = rnd() % 2 + 2;
        vector<array<int, 2>> ch(n, {-1, -1});
        vector<int> sh(n);
        iota(sh.begin(), sh.end(), 0);
        shuffle(sh.begin() + 1, sh.end(), rnd);
        for (int i = 1; i < n; ++i) {
            while (true) {
                int x = sh[rnd() % i];
                if (ch[x][0] != -1 && ch[x][1] != -1) {
                    continue;
                }
                int f = rnd() % 2;
                if (ch[x][f] == -1) {
                    ch[x][f] = sh[i];
                } else {
                    ch[x][f ^ 1] = sh[i];
                }
                break;
            }
        }
        for (int i = 0; i < q; ++i) {
            queries[i].type = 1 + (rnd() % 2);
            if (queries[i].type == 1) {
                queries[i].l = rnd() % n;
                queries[i].r = rnd() % (n - queries[i].l) + queries[i].l;
                queries[i].x = (rnd() % 3) - 1;
            } else {
                queries[i].l = rnd() % n;
            }
        }
        auto st = stupid(ch);
        auto sm = smart(ch);
        if (st != sm) {
            cout << "FOUND!\n";
            cout << n << " " << q << endl;
            for (int i = 0; i < n; ++i) {
                for (int to : ch[i]) {
                    cout << to + 1 << " ";
                }
                cout << endl;
            }
            for (int i = 0; i < q; ++i) {
                if (queries[i].type == 1) {
                    cout << "1 " << queries[i].l + 1 << " " << queries[i].r + 1 << " " << queries[i].x << endl;
                } else {
                    cout << "2 " << queries[i].l + 1 << endl;
                }
            }
            cout << "smart: ";
            for (int x : sm) {
                cout << x << " ";
            }
            cout << endl;
            cout << "stupid: ";
            for (int x : st) {
                cout << x << " ";
            }
            cout << endl;
            exit(0);
        }
        static int cnt = 0;
        if (++cnt % 100 == 0) {
            cout << "OK: " << cnt << endl;
        }
    }
}

int main() {
    ios::sync_with_stdio(false);
    cin.tie(nullptr);

//    stress();
//    return 0;

    cin >> n >> q;

    vector<array<int, 2>> ch(n);
    vector<int> sz(n), init(n), depth(n);
    for (int i = 0; i < n; ++i) {
        cin >> ch[i][0] >> ch[i][1];
        --ch[i][0], --ch[i][1];
    }

    for (int i = 0; i < q; ++i) {
        cin >> queries[i].type;
        if (queries[i].type == 1) {
            cin >> queries[i].l >> queries[i].r >> queries[i].x;
            --queries[i].l, --queries[i].r;
        } else {
            cin >> queries[i].l;
            --queries[i].l;
        }
    }

    auto st = smart(ch);
    for (int i = 0; i < q; ++i) {
        if (queries[i].type == 2) {
            cout << st[i] << '\n';
        }
    }

    return 0;
}
